#ifndef GUARD_misc_h
#define GUARD_misc_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file misc.h
/// @brief Miscellaneous functions.
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <vector>
#include <map>
#include <set>
#include <string>
#include <cstdio>
#include <cctype>

#include <glibmm.h>

/// @brief Function object to convert char to lower case.
struct ToLower: std::unary_function<char, void>
{
    void operator()(char& p) const
    {
	p = std::tolower(p);
    }
};

/// @brief Function object to convert char to upper case.
struct ToUpper: std::unary_function<char, void>
{
    void operator()(char& p) const
    {
	p = std::toupper(p);
    }
};

/// @brief Function object to test char is white-space character.
struct IsCharSpace: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
	return isspace(p);	
    }
};    

/// @brief Function object to test char is '='.
struct IsCharEqual: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
	if(p == '=')
	    return true;
	else
	    return false;
    }
};

/// @brief Function object to test char is '/'.
struct IsCharSlash: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
	if(p == '/')
	    return true;
	else
	    return false;    
    }
};

/// @brief Function object to test char is '.'.
struct IsCharComa: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
	if(p == '.')
	    return true;
	else
	    return false;    
    }
};

/// @brief Test string is just white-space.
inline bool StringIsSpace(const Glib::ustring& str)
{
    return find_if(str.begin(), str.end(), std::not1(IsCharSpace())) == str.end();
}

/// @brief Test path is directory.
inline bool IsDirectory(const std::string& p)
{
    return Glib::file_test(p, Glib::FILE_TEST_IS_DIR);
}

/// @brief Test path is executable.
///
/// See also Glib::file_test
inline bool IsExecutable(const std::string& p)
{
    return (access(p.c_str(), X_OK) == 0);
}

/// @brief Test path is readable.
inline bool IsReadable(const std::string& p)
{
    return (access(p.c_str(), R_OK) == 0);
}

/// @brief Get extention part of filename.
inline std::string GetExt(const std::string& filename)
{
    std::string::const_reverse_iterator coma = find_if(filename.rbegin(), filename.rend(), IsCharComa());
    if(coma != filename.rend()) {
	std::string::const_iterator space = find_if(coma.base(), filename.end(), IsCharSpace());
	std::string temp(coma.base(), space);
	for_each(temp.begin(), temp.end(), ToLower());

	return temp;
    } else
	return std::string();
}

/// @brief Microseconds interval sleep.
///
/// This use select function.
inline int usecsleep(const int& secs, const int& usecs)
{
    struct timeval tv;

    tv.tv_sec = secs;
    tv.tv_usec = usecs;

    return select(0, NULL, NULL, NULL, &tv);
}

inline Glib::ustring ConvertSecondToTimeString(const unsigned long& second)
{
    char timeStr[18];
    std::snprintf(timeStr, sizeof(timeStr), "%02lu:%02lu", second / 60, second % 60);
    
    return timeStr;
}

Glib::ustring StringToUTF8(const Glib::ustring& orig, const std::string& fail, const std::string& assume = "eucKR");

void SplitEachDirToVector(const std::string& fullpath, std::vector<std::string>& dest);

void GetFilenameFromDirToVector(const std::string& path, std::vector<std::string>& dest, bool recursive = true);

bool ConfigfileToMap(const std::string& filename, std::map<std::string, std::string>& configMap, const bool& convertLowerCase = true, const bool& valueBreakAtSpace= true);

#endif

