#ifndef GUARD_player_aac_h
#define GUARD_player_aac_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_aac.h
/// @brief Class for mpeg 4 audio decoding using FAAD2 header file.
/// @todo Deal mp4ff.h more smartly.

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cstdio>
#include <string>

#include <glibmm.h>

#include <faad.h>

// Unix like machine doesn't need mp4ff_int_type.h maybe.
// But mp4ff.h includes mp4ff_int_type.h. So check it and
// include mp4ff.h manually.
#ifdef HAVE_MP4FF_INT_TYPES_H
#include <mp4ff.h>
#else
// from FAAD2 common/mp4ff/mp4ff.h
extern "C" {

    typedef struct
    {
	uint32_t (*read)(void *user_data, void *buffer, uint32_t length);
	uint32_t (*write)(void *udata, void *buffer, uint32_t length);
	uint32_t (*seek)(void *user_data, uint64_t position);
	uint32_t (*truncate)(void *user_data);
	void *user_data;
    } mp4ff_callback_t;

    typedef void* mp4ff_t;

    mp4ff_t *mp4ff_open_read(mp4ff_callback_t *f);
    void mp4ff_close(mp4ff_t *f);

    int32_t mp4ff_get_sample_duration(const mp4ff_t *f, const int32_t track, const int32_t sample);
    int32_t mp4ff_get_sample_duration_use_offsets(const mp4ff_t *f, const int32_t track, const int32_t sample);
    int64_t mp4ff_get_sample_position(const mp4ff_t *f, const int32_t track, const int32_t sample);
    int32_t mp4ff_get_sample_offset(const mp4ff_t *f, const int32_t track, const int32_t sample);

    int32_t mp4ff_find_sample(const mp4ff_t *f, const int32_t track, const int64_t offset,int32_t * toskip);
    int32_t mp4ff_find_sample_use_offsets(const mp4ff_t *f, const int32_t track, const int64_t offset,int32_t * toskip);

    int32_t mp4ff_read_sample(mp4ff_t *f, const int track, const int sample,
	    unsigned char **audio_buffer,  unsigned int *bytes);
    int32_t mp4ff_read_sample_v2(mp4ff_t *f, const int track, const int sample,unsigned char *buffer);
    int32_t mp4ff_read_sample_getsize(mp4ff_t *f, const int track, const int sample);


    int32_t mp4ff_get_decoder_config(const mp4ff_t *f, const int track,
	    unsigned char** ppBuf, unsigned int* pBufSize);
    int32_t mp4ff_get_track_type(const mp4ff_t *f, const int track);
    int32_t mp4ff_total_tracks(const mp4ff_t *f);
    int32_t mp4ff_num_samples(const mp4ff_t *f, const int track);
    int32_t mp4ff_time_scale(const mp4ff_t *f, const int track);

    uint32_t mp4ff_get_avg_bitrate(const mp4ff_t *f, const int32_t track);
    uint32_t mp4ff_get_max_bitrate(const mp4ff_t *f, const int32_t track);
    int64_t mp4ff_get_track_duration(const mp4ff_t *f, const int32_t track);
    int64_t mp4ff_get_track_duration_use_offsets(const mp4ff_t *f, const int32_t track);
    uint32_t mp4ff_get_sample_rate(const mp4ff_t *f, const int32_t track);
    uint32_t mp4ff_get_channel_count(const mp4ff_t * f,const int32_t track);
    uint32_t mp4ff_get_audio_type(const mp4ff_t * f,const int32_t track);


    int mp4ff_meta_get_num_items(const mp4ff_t *f);
    int mp4ff_meta_get_by_index(const mp4ff_t *f, unsigned int index,
	    char **item, char **value);
    int mp4ff_meta_get_title(const mp4ff_t *f, char **value);
    int mp4ff_meta_get_artist(const mp4ff_t *f, char **value);
    int mp4ff_meta_get_writer(const mp4ff_t *f, char **value);
    int mp4ff_meta_get_album(const mp4ff_t *f, char **value);
    int mp4ff_meta_get_date(const mp4ff_t *f, char **value);
    int mp4ff_meta_get_tool(const mp4ff_t *f, char **value);
    int mp4ff_meta_get_comment(const mp4ff_t *f, char **value);
    int mp4ff_meta_get_genre(const mp4ff_t *f, char **value);
    int mp4ff_meta_get_track(const mp4ff_t *f, char **value);
    int mp4ff_meta_get_disc(const mp4ff_t *f, char **value);
    int mp4ff_meta_get_compilation(const mp4ff_t *f, char **value);
    int mp4ff_meta_get_tempo(const mp4ff_t *f, char **value);
    int32_t mp4ff_meta_get_coverart(const mp4ff_t *f, char **value);

    typedef struct
    {
	char *item;
	char *value;
    } mp4ff_tag_t;

    typedef struct
    {
	mp4ff_tag_t *tags;
	uint32_t count;
    } mp4ff_metadata_t;

    int32_t mp4ff_meta_update(mp4ff_callback_t *f,const mp4ff_metadata_t * data);

}
#endif

#include "player.h"

class MainManager;

namespace FAAD {

    class aac_buffer {
	public:
	    aac_buffer()
		: bytes_into_buffer(0),
		bytes_consumed(0),
		file_offset(0),
		at_eof(0),
		buffer(NULL),
		infile(NULL)
		{}

	    ~aac_buffer()
	    {
		if(buffer != NULL) {
		    free(buffer);
		    buffer = NULL;
		}
	    }

	    long bytes_into_buffer;
	    long bytes_consumed;
	    long file_offset;
	    int at_eof;
	    unsigned char* buffer;
	    FILE* infile;
    };

    /// @brief Class for mpeg 4 audio decoding using FAAD2.
    /// @todo Code clean up and real testing.
    class AACPlayer : public Player
    {
	public:
	    AACPlayer(MainManager& mainManager);	    
	    virtual ~AACPlayer();

	    virtual bool Init(const std::string& filename);
	    virtual void Start();
	    virtual void Stop();
	    virtual void Seek(const unsigned long& seconds);
	    virtual const long int& CurrentTime() const;

	    static bool IsMP4File(FILE* const inFile);
	    static int GetAACTrack(mp4ff_t* const inFile);
	    static void AACFillBuffer(aac_buffer* const aacBuffer);
	    static void AACAdvanceBuffer(aac_buffer* const aacBuffer, const int& bytes);

	private:
	    AACPlayer(const AACPlayer& aacPlayer);
	    AACPlayer& operator=(const AACPlayer& aacPlayer);

	    std::string filename_;

	    Glib::Dispatcher signalFinish_;
	    Glib::Dispatcher signalTimerUpdate_;

	    DecodeState decodeState_;

	    long int currentSeconds_;
	    unsigned long seekSeconds_;

	    bool mp4File_;

	    MainManager& mainManager_;

	    virtual void Clean();

	    void MP4Decode();
	    void AACDecode();
    };
}

uint32_t MP4ReadCallback(void* userData, void* buffer, uint32_t length);
uint32_t MP4SeekCallback(void* userData, uint64_t position);

std::auto_ptr<Player> CreateAACPlayer(MainManager& mainManager);

#endif

