#ifndef GUARD_player_mp3_h
#define GUARD_player_mp3_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_mp3.h
/// @brief Class for play mp3 header file.

#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>

#include <vector>
#include <string>

#include <glibmm.h>

#include <mad.h>
#include "player.h"

class MainManager;

namespace MAD {

    struct buffer {
	unsigned char* fdm;
	unsigned char* data;
	unsigned long length;

	int fd;
	int eof;

	struct stat stat;
    };

    struct Timer {
	mad_timer_t playTimer_;
	unsigned long playFrameCount_;
	long playTimeSeconds_;
    };

    /// @brief Class for play mp3.
    ///
    /// MAD requires static function, almost member functions and variables are static.
    class MP3Player : public Player 
    {
	public:
	    MP3Player(MainManager& mainManager);
	    virtual ~MP3Player();

	    virtual bool Init(const std::string& filename);

	    virtual void Start();

	    virtual void Stop(); 
	    
	    virtual void Seek(const unsigned long& seconds);

	    /// @brief Get current time.
	    /// @return current decoded time. 
	    virtual const long int& CurrentTime() const
	    { 
		return timer_().playTimeSeconds_;
	    }

	private:
	    MP3Player& operator=(const MP3Player& mp3Player);
	    MP3Player(const MP3Player& mp3Player);
	
	    static MainManager* mainManager_;		///< Program control MainManager
	    
	    Glib::Dispatcher signalFinish_;		///< Dispatcher for finish signal
	    static Glib::Dispatcher signalTimerUpdate_;	///< Dispatcher for time update signal

	    static std::vector<double> bufSample_;	///< Container to store decoded samples

	    /// @brief String variable for filename.
	    ///
	    /// To declare in static variable, this forms meyers' singleton.
	    /// @return static string variable
	    static std::string& filename_()
	    {
		static std::string filename__;

		return filename__;
	    }

	    struct mad_decoder decoder_;
	    struct buffer buf_;	

	    /// @brief DecodeState variable to check status.
	    ///
	    /// DecodeState is in Player. To declare in static variable,
	    /// this forms meyers' singleton.
	    /// @return static string variable
	    static DecodeState& decodeState_()
	    {
		static DecodeState decodeState__;

		return decodeState__;
	    }

	    /// @brief Integer variable for sample rate.
	    ///
	    /// To declare in static variable, this forms meyers' singleton.
	    /// @return static string variable
	    static unsigned int& sampleRate_()
	    {
		static unsigned int sampleRate__;

		return sampleRate__;
	    }

	    /// @brief Timer variable for main timer.
	    ///
	    /// To declare in static variable, this forms meyers' singleton.
	    /// @return static string variable
	    static Timer& timer_()
	    {
		static Timer timer__;

		return timer__;
	    }

	    /// @brief Integer variable to store wanted seek seconds.
	    ///
	    /// To declare in static variable, this forms meyers' singleton.
	    /// @return static string variable
	    static long& seekSeconds_()
	    {
		static long seekSeconds__;

		return seekSeconds__;
	    }

	    /// @brief Clean up decoder.
	    virtual void Clean()
	    {
	    	mad_decoder_finish(&decoder_);

		if(buf_.fdm)
		    munmap(buf_.fdm, buf_.stat.st_size);

		buf_.fdm = 0;
		
		if(!buf_.eof)
		    buf_.data = 0;    

		if(buf_.data) {
		    free(buf_.data);
		    buf_.data = 0;
		}
		
		if(buf_.fd != -1)
		    close(buf_.fd);
	    }
	    
	    void Reset();

	    static enum mad_flow Input(void* data, struct mad_stream* stream);
	    static enum mad_flow Output(void* data, struct mad_header const* header, struct mad_pcm* pcm);
	    static enum mad_flow Header(void* data, struct mad_header const* header);
	    static enum mad_flow Error(void* data, struct mad_stream* stream, struct mad_frame* frame); 
    };

}

std::auto_ptr<Player> CreateMP3Player(MainManager& mainManager);

#endif

