#ifndef GUARD_playlist_h
#define GUARD_playlist_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file playlist.h
/// @brief Class for Playlist header file.

#include <string>
#include <queue>

#include <gtkmm/treeview.h>
#include <gtkmm/treemodel.h>
#include <gtkmm/liststore.h>
#include <gtkmm/menu.h>

class MainWnd;
class Tag;
class TagFactory;

/// @brief Class for Playlist.
/// @todo Reduce relationship to MainWnd.
class Playlist : public Gtk::TreeView
{
    public:
	Playlist(MainWnd& mainWnd, const std::string& listId, const unsigned int& titleWidth = 470, const bool& loadlist = true);
	virtual ~Playlist();

	void AddFile(const std::string& filePath, const bool& addListFile);
	void AddFiles(const std::vector<std::string>& files, const bool& addListFile);
	void AddDirectory(const std::string& directoryPath, const bool& recursive);
	
	/// @brief Remove all items.
	void RemoveAll() 
	{   
	    OnMenuRemoveAll(); 
	}
	
	std::string Get();
	std::string GetNext();
	std::string GetPrev();
	std::string GetRandom();
	void ForceUpdateToppage();
	void ForceUpdatePage();

	void ShowCurrentFileInfoDlg();

    private:
	class ModelColumns : public Gtk::TreeModel::ColumnRecord
	{
	    public:
		ModelColumns()
		{
		    add(tagParsed_);
		    add(id_);
		    add(filename_);
		    add(title_);
		    add(time_);
		    add(timeSeconds_);
		}

		Gtk::TreeModelColumn<bool> tagParsed_;
		Gtk::TreeModelColumn<unsigned int> id_;
		Gtk::TreeModelColumn<std::string> filename_;
		Gtk::TreeModelColumn<Glib::ustring> title_;
		Gtk::TreeModelColumn<Glib::ustring> time_;
		Gtk::TreeModelColumn<unsigned int> timeSeconds_;
	};

	struct EachRow 
	{
	    Gtk::TreeModel::iterator iter_;
	    bool tagParsed_;
	    Glib::ustring title_;
	    Glib::ustring time_;
	    unsigned int timeSeconds_;
	};

	const std::string fileProtocol_;		///< String to determine file protocol, "file://"
	std::string listId_;				///< Identification string of list.
	std::string dragedFilename_;			///< Temporary string to store string in DragNDrop.

	ModelColumns columns_;

	Gtk::Menu menuPopup_;
	Gtk::TreeModel::iterator pos_;

	MainWnd& mainWnd_;

	Glib::Thread* tagParsingThread_;
	Glib::Mutex tagParsingMutex_;
	Glib::Dispatcher signalRowToFill_;
	std::queue<EachRow> rowQueue_;
	void RowFill(); 

	unsigned int count_;				///< Number of added items.
	unsigned int tagProcessCount_;			///< Number of items reading tag at one thread.

	bool inTagParsing_;				///< Tag parsing thread is running.
	bool stopTagParsing_;				///< To stop tag parsing.

	Glib::RefPtr<Gtk::ListStore> treeModel_;
	
	std::auto_ptr<TagFactory> tagFactory_;
	
	virtual bool on_button_press_event(GdkEventButton* ev);
	virtual void on_row_activated(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column);
	virtual void on_set_scroll_adjustments(Gtk::Adjustment* hadj, Gtk::Adjustment* vadj);
	virtual void on_size_allocate(Gtk::Allocation& allocation);

	/// @brief Filename compare function to sort list.
	int FilenameCompare(const Gtk::TreeModel::iterator& a, const Gtk::TreeModel::iterator& b)
	{
	    std::string strA = (*a)[columns_.filename_];
	    std::string strB = (*b)[columns_.filename_];

	    std::string aPath = Glib::path_get_dirname(strA);
	    std::string bPath = Glib::path_get_dirname(strB);
	    
	    // If files in same directory, sort them
	    // or sort for directory.	
	    if(aPath == bPath)
		return strcmp(strA.c_str(), strB.c_str());
	    else
		return strcmp(aPath.c_str(), bPath.c_str());
	}

	/// @brief Move cursor to coresspond iterator and scroll to that.
	void MoveCursorAndScroll(const Gtk::TreeModel::iterator& iter)
	{
	    Gtk::TreeModel::Path path;
	    if(iter) {
		path = Gtk::TreeModel::Path(iter);
	    } else {
		path = Gtk::TreeModel::Path("0");
	    }
	    set_cursor(path);
	    scroll_to_row(path, 0.5);
	}

	std::string GetFilenameFromTreeiter(Gtk::TreeModel::iterator& iter); 
	void ParseTag(Gtk::TreeModel::iterator& iter, const bool& thread);
	void GetInfoFromTag(Gtk::TreeModel::iterator& iter);

	void ShowFileInfoDlg(const std::string& filename);

	void OnDragDataReceived(const Glib::RefPtr<Gdk::DragContext>& context, int x, int y, const Gtk::SelectionData& selectionData, guint info, guint time);

	void OnMenuRemove();
	void OnMenuRemoveAll();
	void OnMenuSortFilename();
	void OnMenuInfo();

	void OnVScrolled();

	void LoadLastList();
    
	void LoadList(const std::string& filename);
	void SaveList();

	void StopThread();

};

#endif

