// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_ogg.cc
/// @brief Class for play ogg source file.

#include "../include/player_ogg.h"
#include "../include/mainmanager.h"
#include "../include/trace.h"

/// @brief Constructor of OGGPlayer.
///
/// Connect dispatchers to functions.
VORBIS::OGGPlayer::OGGPlayer(MainManager& mainManager)
: decodeState_(STOP),
currentSeconds_(0),
buflen_(4096),
sampleRate_(0),
ret_(0),
bs_(0),
ovOpenSuccess_(true),
mainManager_(mainManager),
inFile_(NULL)
{
    signalFinish_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayFinish));
    signalTimerUpdate_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayTimeUpdate));
}

VORBIS::OGGPlayer::~OGGPlayer()
{
    Clean();
}

/// @brief Initialize OGGPlayer.
///
/// Just open file in this function. ov_open is called in Start().
/// @param filename filename to play.
/// @return Success of file open.
bool VORBIS::OGGPlayer::Init(const std::string& filename)
{
    Tracer::Trace trace("VORBIS::OGGPlayer::Init");

    inFile_ = fopen(filename.c_str(), "rb");
    
    if(!inFile_) {
	trace.Debug("Can't open file");
	 
	return false;
    }

    return true;
}

/// @brief Start player loop.
///
/// Decode file and send decoded samples to play function. Timer update and
/// audio open, close is also in this function.
/// @todo Avoid infinite loop problem on EOF and OV_HOLE in sometimes.
void VORBIS::OGGPlayer::Start()
{
    Tracer::Trace trace("VORBIS::OGGPlayer::Start");

    if(ov_open(inFile_, &vorbisFile_, NULL, 0) < 0) {
	trace.Debug("Error at ov_open");
		
	ovOpenSuccess_ = false;	
	if(fclose(inFile_) != 0)
	    trace.Debug("fclose failed");
	
	inFile_ = NULL;    
	
	signalFinish_();
	return;
    }

    decodeState_ = PLAY;

    for(;;) {	

	if(decodeState_ == STOP) {
	    trace.Debug("decodeState_ == STOP");
	    break;
	}

	
	{
	    Glib::Mutex::Lock lock(vfMutex_);	
	    
	    ret_ = ov_read_float(&vorbisFile_, &buf_, buflen_, &bs_);
	}


	if(ret_ == 0) {
	    trace.Debug("End of File");
	    break;
	    
	} else if(ret_ == OV_EBADLINK) {
	    trace.Debug("Got error, invalid stream section was supplied, or the requested link is corrupt(OV_EBADLINK)");

	} else if(ret_ == OV_HOLE) {
	    trace.Debug("Gor error, interruption in the data(OV_HOLE)");
	    
	} else if(ret_ < 0) {
	    trace.Debug("Got error ret_ < 0");    
	    
	} else if(ret_ > 0) {

	    if(bs_ != 0) {
		trace.Debug("bs_ != 0");
		break;
	    }

	    if(!mainManager_.AudioCheckOpenDevice() || sampleRate_ != vorbisFile_.vi->rate) {
		sampleRate_ = vorbisFile_.vi->rate;
		mainManager_.AudioClose(true);

		trace.Debug("OpenDevice");
		mainManager_.AudioOpenDevice(16, vorbisFile_.vi->channels, sampleRate_, mainManager_.audioFormatNativeEndian_);	
	    }

	    mainManager_.AudioPlay(buf_, ret_);

	    long int seconds;
	
	    {
		Glib::Mutex::Lock lock(vfMutex_);
	    
		seconds = static_cast<long int>(ov_time_tell(&vorbisFile_));
	    }
	
	    if( currentSeconds_ != seconds ) {
	    
		currentSeconds_ = seconds;
	    
		signalTimerUpdate_();
	    
	    }

	} 

    }
    
    if(decodeState_ == PLAY)
	mainManager_.AudioClose(true);
    else
	mainManager_.AudioClose(false);
		
    signalFinish_();

}

/// @brief Set decodeState to STOP.
///
/// Set decodeState to STOP. Detect STOP in Start() function.
void VORBIS::OGGPlayer::Stop()
{
    decodeState_ = STOP;
}

/// @brief Seek to wanted time.
/// @param seconds wanted time to seek in seconds.
void VORBIS::OGGPlayer::Seek(const unsigned long& seconds)
{
    {
	Glib::Mutex::Lock lock(vfMutex_);
	
	ov_time_seek(&vorbisFile_, seconds);
    }
}

/// @brief Get current decoded time.
/// @return current decoded time in seconds.
const long int& VORBIS::OGGPlayer::CurrentTime() const
{
    return currentSeconds_;
}

/// @brief Clean up decoder.
void VORBIS::OGGPlayer::Clean()
{
    Tracer::Trace trace("VORBIS::OGGPlayer::Clean");
    
    // NOTICE:
    //	According to vorbis api documentation, if you success on ov_open,
    // you should use ov_clear and not use fclose. But you fail on ov_open,
    // you should use fclose to close a FILE pointer.

    if(ovOpenSuccess_) {	
	trace.Debug("ov_clear");
	ov_clear(&vorbisFile_);
    } else if(inFile_ != NULL) {
	if(fclose(inFile_) != 0)
	    trace.Debug("fclose failed");
	    
	inFile_ = NULL;    
    }
    
}

/// @brief Create OGGPlayer object.
///
/// Used in PlayerFactory.
std::auto_ptr<Player> CreateOGGPlayer(MainManager& mainManager)
{
    return std::auto_ptr<Player>(new VORBIS::OGGPlayer(mainManager));
}

