// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file playlist.cc
/// @brief Class for Playlist source file.

#include <iostream>
#include <fstream>

#include <vector>
#include <list>
#include <map>

#include <gdkmm.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <libintl.h>

#include "../include/playlist.h"
#include "../include/mainwin.h"
#include "../include/tag_mp3.h"
#include "../include/tag_ogg.h"
#include "../include/tag_wma.h"

#ifdef ENABLE_FAAD2
#include "../include/tag_aac.h"
#endif

#include "../include/fileinfodlg.h"
#include "../include/trace.h"
#include "../include/misc.h"

/// @brief Constructor of Playlist.
/// @param mainWnd MainWnd
/// @param listId Identification string of list.
/// @param titleWidth Title column width.
/// @param loadlist Load list from listId file at first.
Playlist::Playlist(MainWnd& mainWnd, const std::string& listId, const unsigned int& titleWidth, const bool& loadlist)
: fileProtocol_("file://"),
listId_(listId),
mainWnd_(mainWnd),
tagParsingThread_(NULL),
count_(0),
tagProcessCount_(0),
inTagParsing_(false),
stopTagParsing_(false)
{
    #warning "Fix this magic constant!"
    tagProcessCount_ = (mainWnd_.prevHeight_ - 170) / 25 + 2;
    
    treeModel_ = Gtk::ListStore::create(columns_);
    set_model(treeModel_);

    append_column(gettext("Number"), columns_.id_);
    append_column(gettext("Title"), columns_.title_);
    append_column(gettext("Time"), columns_.time_);

    Gtk::Menu_Helpers::MenuList& popupMenuList = menuPopup_.items();
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Add files"),
	sigc::mem_fun(mainWnd_, &MainWnd::OnMenuFileAdd) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Add a _directory"),
	sigc::mem_fun(mainWnd_, &MainWnd::OnMenuFileDirAdd) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Remove"),
	sigc::mem_fun(*this, &Playlist::OnMenuRemove) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Re_move all"),
	sigc::mem_fun(*this, &Playlist::OnMenuRemoveAll) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Sort filename"),
	sigc::mem_fun(*this, &Playlist::OnMenuSortFilename) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Information"),
	sigc::mem_fun(*this, &Playlist::OnMenuInfo) ));

    // Column 1 is title.
    get_column(1)->set_resizable(true);
    get_column(1)->set_fixed_width(titleWidth);
    get_column(1)->set_sizing(Gtk::TREE_VIEW_COLUMN_FIXED);
    
    tagFactory_ = std::auto_ptr<TagFactory>(new TagFactory);
    tagFactory_->Register("mp1", &mainWnd_.mainManager_, &CreateId3Tag);
    tagFactory_->Register("mp2", &mainWnd_.mainManager_, &CreateId3Tag);
    tagFactory_->Register("mp3", &mainWnd_.mainManager_, &CreateId3Tag);
    tagFactory_->Register("ogg", &mainWnd_.mainManager_, &CreateOGGTag);
    tagFactory_->Register("wma", &mainWnd_.mainManager_, &CreateWMATag);

#ifdef ENABLE_FAAD2
    tagFactory_->Register("mp4", &mainWnd_.mainManager_, &CreateAACTag);
    tagFactory_->Register("m4a", &mainWnd_.mainManager_, &CreateAACTag);
    tagFactory_->Register("aac", &mainWnd_.mainManager_, &CreateAACTag);
#endif    

    if(loadlist)
	LoadLastList();

    treeModel_->set_sort_func(columns_.filename_, 
    	sigc::mem_fun(*this, &Playlist::FilenameCompare));

    signalRowToFill_.connect(sigc::mem_fun(*this, &Playlist::RowFill));

    std::list<Gtk::TargetEntry> targets;
    targets.push_back( Gtk::TargetEntry("text/plain") );
    drag_dest_set(targets);
    signal_drag_data_received().connect(sigc::mem_fun(*this, &Playlist::OnDragDataReceived));

    ForceUpdateToppage();
}

Playlist::~Playlist()
{
    StopThread();
    
    SaveList();
}

/// @brief Add file to list.
///
/// If file is not playable and file is not list-type file, 
/// file will be rejected.
/// @param filePath A filepath to add.
/// @param addListFile Load listfile if filepath is list-type file.
/// @todo Fix magic constant for list-type. 
void Playlist::AddFile(const std::string& filePath, const bool& addListFile)
{
    if(mainWnd_.IsPlayable(filePath)) {

	Gtk::TreeModel::Row row = *(treeModel_->append());
	row[columns_.tagParsed_] = false;
	row[columns_.id_] = ++count_;
	row[columns_.filename_] = filePath;
	row[columns_.timeSeconds_] = 1;
	
    } else if(addListFile && GetExt(filePath) == "m3u") 
	LoadList(filePath);
}

/// @brief Add files to list.
/// @param files Files to add. 
/// @param addListFile Load listfile if filepath is list-type file.
void Playlist::AddFiles(const std::vector<std::string>& files, const bool& addListFile)
{
    typedef std::vector<std::string>::const_iterator iter;
    
    iter end = files.end();
    for(iter eachFile = files.begin(); eachFile != end; ++eachFile) {
	
	AddFile(*eachFile, addListFile);
	
    }
}

/// @brief Add files under directory to playlist.
/// @param directoryPath Directory to add.
/// @param recursive Read directory recursively or not.
void Playlist::AddDirectory(const std::string& directoryPath, const bool& recursive)
{
    std::vector<std::string> filesToAdd;
    GetFilenameFromDirToVector(directoryPath, filesToAdd, recursive);

    AddFiles(filesToAdd, false);
}

/// @brief Get file path.
std::string Playlist::Get()
{
    return GetFilenameFromTreeiter(pos_);
}

/// @brief Get next position item's file path.
std::string Playlist::GetNext()
{
    if(pos_) {
	
	if( pos_ == --(treeModel_->children().end()) ) {
	   
	    if(!mainWnd_.repeatPlay_) {

		mainWnd_.listState_ = MainWnd::LISTSTATE_LAST;
			
		return std::string();
	    }
	}

	return GetFilenameFromTreeiter(++pos_);	
	
    } else
	return GetFilenameFromTreeiter(pos_);	
}

/// @brief Get previous position item's file path.
std::string Playlist::GetPrev()
{
    if(pos_) {

	if(pos_ == treeModel_->children().begin()) {
	    
	    if(mainWnd_.repeatPlay_)
		pos_ = treeModel_->children().end();
	    else {	

		mainWnd_.listState_ = MainWnd::LISTSTATE_FIRST;

		return std::string();
	    }
	}
	
	return GetFilenameFromTreeiter(--pos_);	
	
    } else
	return GetFilenameFromTreeiter(pos_);
}

/// @brief Get item's file path from random position.
std::string Playlist::GetRandom()
{
    if(count_ != 0) {
	
	Glib::Rand rand;

	int randomNumber = rand.get_int_range(0, count_);

	pos_ = treeModel_->children().begin();
    
	for(int i = 0; i < randomNumber; i++)
	    pos_++;

	return GetFilenameFromTreeiter(pos_);
	
    } else {
	mainWnd_.listState_ = MainWnd::LISTSTATE_EMPTY;
		
	return std::string();
    }
}

/// @brief Force tag parsing from top page.
void Playlist::ForceUpdateToppage()
{
    Gtk::TreeModel::iterator iter = treeModel_->children().begin();

    if(iter) {
	StopThread();
	inTagParsing_ = true;
	tagParsingThread_ = Glib::Thread::create( sigc::bind(
	   sigc::mem_fun(*this, &Playlist::GetInfoFromTag), iter), true );
    }
}

/// @brief Force tag parsing from current view page.
void Playlist::ForceUpdatePage()
{
    while(Glib::MainContext::get_default()->iteration(false));
    OnVScrolled();
}

/// @brief Show current selected item's information.
void Playlist::ShowCurrentFileInfoDlg()
{
    if(pos_)
	ShowFileInfoDlg( (*pos_)[columns_.filename_] );
}

/// @brief Callback function used when button is press.
///
/// If pressed button is 3 button, pop up menu.
bool Playlist::on_button_press_event(GdkEventButton* ev)
{
    bool rtn = TreeView::on_button_press_event(ev);

    if( (ev->type == GDK_BUTTON_PRESS) && (ev->button == 3) ) {
	menuPopup_.popup(ev->button, ev->time);
    }

    return rtn;
}

/// @brief Callback function used when row is activated(double clicked or else ...)
void Playlist::on_row_activated(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column)
{
    Gtk::TreeView::on_row_activated(path, column);

    if(Gtk::TreeModel::iterator iter = treeModel_->get_iter(path)) {
	pos_ = iter;
	
	mainWnd_.selectedFile_ = GetFilenameFromTreeiter(pos_);
    }

    mainWnd_.OnBtnPlay();

}

/// @brief Callback function used when scroll adjustments is set.
///
/// This function called when TreeView regist hadj, vadj and unregist them.
/// OnVScrolled is connected vadj's signal_value_changed().
void Playlist::on_set_scroll_adjustments(Gtk::Adjustment* hadj, Gtk::Adjustment* vadj)
{
    Gtk::TreeView::on_set_scroll_adjustments(hadj, vadj);

    if(vadj) {
	vadj->signal_value_changed().connect(
	    sigc::mem_fun(*this, &Playlist::OnVScrolled) );
    }

}

/// @brief Callback function used when size allocate.
///
/// This function called when widget is added to other container. 
/// tagProcessCount_ is calculated in here.
void Playlist::on_size_allocate(Gtk::Allocation& allocation)
{
    Gtk::TreeView::on_size_allocate(allocation);

    // some magic constant 25
    tagProcessCount_ = allocation.get_height() / 25 + 2;
}

/// @brief Get file path from corressponding iterator.
///
/// Parse tag of file, move cursor and scroll to it.
std::string Playlist::GetFilenameFromTreeiter(Gtk::TreeModel::iterator& iter)
{
    if(!iter) {
	// You changed reference variable!
	iter = treeModel_->children().begin();
	
	if(!iter) {
	    mainWnd_.listState_ = MainWnd::LISTSTATE_EMPTY;
	    
	    return std::string();
	}
    }

    // Mutex is locked in scope, and die at out of scope
    {
	Glib::Mutex::Lock lock(tagParsingMutex_);
	    
	if( !(*iter)[columns_.tagParsed_] ) { 
	    ParseTag(iter, false);
	}
    }

    mainWnd_.DisplayTitle( (*iter)[columns_.title_] );
    mainWnd_.DisplayTotalTime( (*iter)[columns_.time_] );
    mainWnd_.SetSeekScaleRange( (*iter)[columns_.timeSeconds_] );

    MoveCursorAndScroll(iter);

    mainWnd_.listState_ = MainWnd::LISTSTATE_NORMAL;
	
    return std::string( (*iter)[columns_.filename_] );

}

/// @brief Parse tag of file.
/// @param iter Iterator.
/// @param thread If function is called in thread, set true.
void Playlist::ParseTag(Gtk::TreeModel::iterator& iter, const bool& thread)
{
    std::string filename = (*iter)[columns_.filename_];
    
    if(IsReadable(filename)) {
	std::auto_ptr<Tag> tagParser = tagFactory_->Create(filename);

	if(tagParser.get() != NULL) { 
	    if(thread) {
		EachRow thisRow;
		
		thisRow.iter_ = iter;
		thisRow.title_ = tagParser->GetTitle();
		thisRow.time_ = tagParser->GetTotalTime();
		thisRow.timeSeconds_ = tagParser->GetTotalTimeSeconds();

		if(thisRow.timeSeconds_ != 1)
		    thisRow.tagParsed_ = true;
		
		// already locked ...
		rowQueue_.push(thisRow);

		signalRowToFill_();

	    } else {
		(*iter)[columns_.title_] = tagParser->GetTitle();
		(*iter)[columns_.time_] = tagParser->GetTotalTime();
		(*iter)[columns_.timeSeconds_] = tagParser->GetTotalTimeSeconds();

		if((*iter)[columns_.timeSeconds_] != 1)
		    (*iter)[columns_.tagParsed_] = true;

	    }
	}
    }
    
}

/// @brief Fill row from rowQueue.
///
/// This function must be called in thread.
void Playlist::RowFill()
{
    std::auto_ptr<EachRow> processRow;
    
    {
	Glib::Mutex::Lock lock(tagParsingMutex_);

	if(rowQueue_.empty())
	    return;
	
	processRow = std::auto_ptr<EachRow>(new EachRow(rowQueue_.front()));
	rowQueue_.pop();
    }

    (*processRow->iter_)[columns_.tagParsed_] = processRow->tagParsed_;
    (*processRow->iter_)[columns_.title_] = processRow->title_;
    (*processRow->iter_)[columns_.time_] = processRow->time_;
    (*processRow->iter_)[columns_.timeSeconds_] = processRow->timeSeconds_;

}

/// @brief Parse tag loop.
/// @param iter Start iterator
void Playlist::GetInfoFromTag(Gtk::TreeModel::iterator& iter)
{
    unsigned int i = 0;

    while(!stopTagParsing_)
    {
	if(iter) {
	    
	    if(i > tagProcessCount_)
		break; 

	    {
		Glib::Mutex::Lock lock(tagParsingMutex_);
		
		if(!(*iter)[columns_.tagParsed_])
		    ParseTag(iter, true);	
			
	    }
	    
	    iter++;
	    i++;
	} else 
	    break;
    }
}

/// @brief Show information dialog.
void Playlist::ShowFileInfoDlg(const std::string& filename)
{
    if(IsReadable(filename)) {
	std::auto_ptr<Tag> tagParser = tagFactory_->Create(filename);

	if(tagParser.get() != NULL) { 
	    std::map<std::string, Glib::ustring> fileInfoMap;
	
	    fileInfoMap["title"] = tagParser->GetTitle();
	    fileInfoMap["album"] = tagParser->GetAlbum();
	    fileInfoMap["genre"] = tagParser->GetGenre();
	    fileInfoMap["artist"] = tagParser->GetArtist();
	    fileInfoMap["comment"] = tagParser->GetComment();
	    fileInfoMap["year"] = tagParser->GetYear();
	    fileInfoMap["track"] = tagParser->GetTrack();
	
	    FileInfoDlg fileInfoDlg(fileInfoMap);
	    fileInfoDlg.run();
	}
    }
}

/// @brief Callback function used when drag data received.
///
/// When drag data received, add them to list.
void Playlist::OnDragDataReceived(const Glib::RefPtr<Gdk::DragContext>& context, int x, int y, const Gtk::SelectionData& selectionData, guint info, guint time)
{
    if((selectionData.get_length() >=0 ) && (selectionData.get_format() == 8)) {

	dragedFilename_ = selectionData.get_data_as_string();
	std::string::const_iterator startURI = dragedFilename_.begin();
	std::string::const_iterator endURI = dragedFilename_.begin();
	std::string::const_iterator end = dragedFilename_.end();
	
	std::string eachFilename;
	for(;;) {
	    startURI = search(startURI, end, fileProtocol_.begin(), fileProtocol_.end());
	    if(startURI == end)
		break;

	    endURI = find_if(startURI, end, IsCharSpace());

	    if(startURI != endURI) {
		eachFilename = std::string(startURI, endURI);
		try {
		    eachFilename = Glib::filename_from_uri(eachFilename);

		    if(IsDirectory(eachFilename))
			AddDirectory(eachFilename, true);
		    else
			AddFile(eachFilename, true);

		} catch(Glib::ConvertError& err) {
		    
		    // Got convert error!
		}

	    }

	    startURI = endURI;
	}

	ForceUpdateToppage();

    }
    context->drag_finish(false, false, time);
}

/// @brief Callback function used when remove menu is clicked.
void Playlist::OnMenuRemove()
{
    StopThread();
    
    if(Glib::RefPtr<Gtk::TreeView::Selection> selection = get_selection()) {
	if(Gtk::TreeModel::iterator iter = selection->get_selected()) {
	    if(mainWnd_.selectedFile_ == std::string((*iter)[columns_.filename_]))
		mainWnd_.OnBtnStop();
	    
	    unsigned int prevCount = (*iter)[columns_.id_];

	    if(pos_ && pos_ == iter) 
		pos_++;
		
	    iter = treeModel_->erase(iter);	
	    MoveCursorAndScroll(iter);
	    count_--;

	    Gtk::TreeModel::iterator end = treeModel_->children().end();
	    while(iter != end) {
		(*iter)[columns_.id_] = prevCount;		
		prevCount++;
		iter++;
	    }
	}
    }
}

/// @brief Callback function used when remove all menu is clicked.
void Playlist::OnMenuRemoveAll()
{
    StopThread();

    mainWnd_.OnBtnStop();
    mainWnd_.selectedFile_ = "";
    mainWnd_.DisplayTitle("", true);

    Gtk::TreeModel::iterator iter = treeModel_->children().begin();

    while(iter) {
	iter = treeModel_->erase(iter);	
    }

    count_ = 0;
    pos_ = treeModel_->children().begin();
    mainWnd_.selectedFile_ = "";
}

/// @brief Callback function used when sort filename menu is clicked.
void Playlist::OnMenuSortFilename()
{
    treeModel_->set_sort_column(columns_.filename_, Gtk::SORT_ASCENDING);

    ForceUpdatePage();
}

/// @brief Callback function used when information menu is clicked.
void Playlist::OnMenuInfo()
{
    if(Glib::RefPtr<Gtk::TreeView::Selection> selection = get_selection()) {
	if(Gtk::TreeModel::iterator iter = selection->get_selected()) {

	    ShowFileInfoDlg( (*iter)[columns_.filename_] );
	    
	}
    }

}

/// @brief Callback function used when scrolled.
void Playlist::OnVScrolled()
{
    if(count_ == 0)
	return;

    StopThread();
     
    Gtk::TreePath eachPath;
    Gtk::TreeViewColumn* eachColumn;
    int cellx = 0;
    int celly = 0;

    get_path_at_pos(0, 0, eachPath, eachColumn, cellx, celly);	
    
    Gtk::TreeModel::iterator iter = treeModel_->get_iter(eachPath);
    
    inTagParsing_ = true;
    tagParsingThread_ = Glib::Thread::create( sigc::bind(
    	sigc::mem_fun(*this, &Playlist::GetInfoFromTag), iter), true );
}

/// @brief Load list from indentification string file.
void Playlist::LoadLastList()
{
    Tracer::Trace trace("Playlist::LoadLastList");
    
    std::string listFilename = Glib::get_home_dir();
    listFilename += "/.sagua/";
    listFilename += listId_;
    listFilename += ".m3u";

    LoadList(listFilename);
}

/// @brief Load list from filename.
void Playlist::LoadList(const std::string& filename)
{
    Tracer::Trace trace("Playlist::LoadList");
    
    std::string tmp;
    tmp.reserve(50);
    std::ifstream listFile(filename.c_str());
    if(listFile) {
	std::string eachLine;
	std::string::reverse_iterator space;
	while(getline(listFile, eachLine)) {
	    // if line is comment of EXTINF
	    if(eachLine[0] == '#')
		continue;
	    else {
		// To elliminate last space character find space character
		space = find_if(eachLine.rbegin(), eachLine.rend(), std::not1(IsCharSpace()));
		// If find space character mid of string
		if(space != eachLine.rbegin() && space != eachLine.rend())
		    eachLine = std::string(eachLine.begin(), space.base());
		
		// If line has not path, assume that file in list file's dir. 
		if(!Glib::path_is_absolute(eachLine)) {
		    tmp = Glib::path_get_dirname(filename);
		    tmp += '/';
		    tmp += eachLine;
		    AddFile(tmp, false);
		} else
		    AddFile(eachLine, false);
	    } 
	}
    } else
	trace.Debug("Can't open listfile to load");
}

/// @brief Save current list to identification string file.
void Playlist::SaveList()
{
    Tracer::Trace trace("Playlist::SaveList");

    std::string programDir = Glib::get_home_dir();
    programDir += "/.sagua";

    if(!Glib::file_test(programDir, Glib::FILE_TEST_EXISTS)) {	
	trace.Debug("programDir does not exist");
	// make dir with permission 755
	mkdir(programDir.c_str(), S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	trace.Debug("Create programDir");
    } else if(!Glib::file_test(programDir, Glib::FILE_TEST_IS_DIR)) {
	trace.Debug("File exists already in same name");

	return;
    }

    std::string listFilename = programDir;
    listFilename += "/";
    listFilename += listId_;
    listFilename += ".m3u";	

    std::ofstream listFile(listFilename.c_str());

    if(listFile) {
	// m3u format... (is m3u a media queue format? anyway)
	//	#EXTM3U
	//	#EXTINF:song_length_in_seconds,song_title
	//	somefilepath.mp3
	listFile << "#EXTM3U" << std::endl;

	#warning "hey, expection for not parsed file?"
	Gtk::TreeModel::iterator iter = treeModel_->children().begin();

	unsigned int seconds;
	Glib::ustring title;
	std::string filename;

	while(iter) {
	    seconds = (*iter)[columns_.timeSeconds_];
	    title = (*iter)[columns_.title_];
	    filename = (*iter)[columns_.filename_];
	    listFile << "#EXTINF:" << seconds << "," << title << std::endl 
		    << filename << std::endl;
	    iter++;		    
	}
    } else
	trace.Debug("Can't open listFile to write"); 
}

/// @brief Stop thread.
void Playlist::StopThread()
{
    if(inTagParsing_) {
	stopTagParsing_ = true;
	tagParsingThread_->join();
	inTagParsing_ = false;	
	stopTagParsing_ = false;
    }
}

