// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file socket.cc
/// @brief Interface class for unix socket source file

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>

#include "../include/socket.h"

/// @brief Constructor of unix socket.
///
/// Prepare path name for socket and set addr to that path.
UnixSocket::UnixSocket()
: integerSize_(sizeof(int)),
    socketPath_(Glib::get_tmp_dir()),
    fd_(-1),
    inThread_(false),
    stopThread_(false),
    acceptThread_(NULL)
{
    socketPath_ += "/" PACKAGE_NAME "-";
    socketPath_ += Glib::get_user_name();

    bzero(reinterpret_cast<char*>(&servAddr_), sizeof(servAddr_));
    strncpy(servAddr_.sun_path, socketPath_.c_str(), sizeof(servAddr_.sun_path) - 1);

    servAddr_.sun_family = AF_LOCAL;

}

/// @brief Destructor of unix socket.
///
/// Close socket and stop thread
UnixSocket::~UnixSocket()
{
    if(inThread_) {
	stopThread_ = true;
	acceptThread_->join();	
	stopThread_ = false;
	inThread_ = false;
    }

    Close();
}

/// @brief Connect to prepared addr.
///
/// @return Success of connect
bool UnixSocket::Connect()
{
    fd_ = socket(AF_LOCAL, SOCK_STREAM, 0);

    if(fd_ == -1)
	return false;

    if(connect(fd_, reinterpret_cast<struct sockaddr*>(&servAddr_), sizeof(servAddr_)) == -1)
	return false;
    else
	return true;	
}

/// @brief Close socket
void UnixSocket::Close()
{
    if(fd_ != -1)
	close(fd_);
}

/// @brief Listen to connection.
///
/// If there is file in socket path, unlink it and bind and listen.
/// @return Success of listen
bool UnixSocket::Listen()
{
    fd_ = socket(AF_LOCAL, SOCK_STREAM, 0);

    if(fd_ == -1)
	return false;

    UnlinkSocketPath();	

    if(bind(fd_, reinterpret_cast<struct sockaddr*>(&servAddr_), sizeof(servAddr_)) == -1)
	return false;

    if(listen(fd_, 5) == -1)
	return false;
    else
	return true;	

}

/// @brief Send message to socket.
///
/// Send all size of messages to socket. 
/// @param fd Socket file descriptor
/// @param message Message to send
/// @param size Size of message
/// @return Success of send
bool UnixSocket::Send(const int& fd, const void* message, const int& size)
{
    int sentBytes = -1;
    int currentBytes = 0;
    int leftBytes = size;

    while(currentBytes < size) {
	sentBytes = send(fd, reinterpret_cast<const char*>(message) + currentBytes, leftBytes, 0);
	if(sentBytes == -1)
	    return false;

	currentBytes += sentBytes;
	leftBytes -= sentBytes;  
    }

    return true;
}

/// @brief Send message in std::string to socket.
///
/// Send all messages in std::string to socket. 
/// @param fd Socket file descriptor
/// @param message Message to send in std::string
/// @return Success of send
bool UnixSocket::Send(const int& fd, const std::string& message)
{
    int length = message.size();
    int sentBytes = -1;
    int currentBytes = 0;
    int leftBytes = length;

    while(currentBytes < length) {
	sentBytes = send(fd, message.c_str() + currentBytes, leftBytes, 0);
	if(sentBytes == -1)
	    return false;

	currentBytes += sentBytes;
	leftBytes -= sentBytes;  
    }

    return true;
}

/// @brief Send integer 1 to socket
/// @param fd Socket file descriptor
/// @return Success of send
bool UnixSocket::SendSignal(const int& fd)
{
    int buf = 1;

    return Send(fd, &buf, integerSize_);
}

/// @brief Receive from socket and test received data is integer 1
/// @param fd Socket file descriptor
/// @return Success of receive
bool UnixSocket::RecvSignal(const int& fd)
{
    int buf;
    int receivedBytes = recv(fd, &buf, integerSize_, 0);
    if(receivedBytes != -1 && buf == 1)
	return true;
    else
	return false;	
}

/// @brief Run thread for accept and communcation to client
void UnixSocket::Accept()
{
    acceptThread_ = Glib::Thread::create(sigc::mem_fun(*this, &UnixSocket::AcceptLoop), true);
    inThread_ = true;
}

/// @brief Loop for accept and communcation to client
void UnixSocket::AcceptLoop()
{
    char buf[1024];
    fd_set master;
    fd_set readfds;
    struct sockaddr_un clientAddr;
    int clientLen = sizeof(clientAddr);
    int clientFd = -1;
    int fdmax = 0;
    int receivedBytes = -1;

    FD_ZERO(&master);
    FD_ZERO(&readfds);
    FD_SET(fd_, &master);

    fdmax = fd_;

    struct timeval tv;

    while(!stopThread_) {
	tv.tv_sec = 0;
	tv.tv_usec = 1000;

	readfds = master;
	if(select(fdmax + 1, &readfds, NULL, NULL, &tv) == -1)
	    continue;

	for(int i = 0; i <= fdmax; ++i) {
	    if(FD_ISSET(i, &readfds)) {
		if(i == fd_) {
		    clientLen = sizeof(clientAddr);
		    clientFd = accept(fd_, reinterpret_cast<struct sockaddr*>(&clientAddr), reinterpret_cast<socklen_t*>(&clientLen));
		    if(clientFd == -1)
			continue;
		    else {
			FD_SET(clientFd, &master);
			if(clientFd > fdmax)
			    fdmax = clientFd;
		    }
		} else {
		    receivedBytes = -1;
		    receivedBytes = recv(i, buf, 1023, 0);
		    if(receivedBytes > 0)
			DataReceived(i, receivedBytes, buf, 1023);
		    else {
			close(i);
			FD_CLR(i, &master);
		    }
		}
	    }
	}
    }
}

/// @brief If is file in socket path, unlink it.
void UnixSocket::UnlinkSocketPath()
{
    if(Glib::file_test(socketPath_, Glib::FILE_TEST_EXISTS))
	unlink(socketPath_.c_str());
}

