// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file sockethandler.cc
/// @brief Class for unix socket handler source file.

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "../include/sockethandler.h"
#include "../include/mainmanager.h"

/// @brief Constructor of UnixSocketHandler.
/// @param arguments Command line arguments.
UnixSocketHandler::UnixSocketHandler(const std::vector<std::string>& arguments)
: arguments_(arguments),
mainManager_(NULL)
{}

UnixSocketHandler::~UnixSocketHandler()
{}

/// @brief Send command line arguments to server.
void UnixSocketHandler::SendArguments()
{
    if(!arguments_.empty()) {
	int totalBytes = 0;
	int thisBytes = 0;
	size_t integerSize = sizeof(totalBytes);
	for(std::vector<std::string>::const_iterator iter = arguments_.begin();
		iter != arguments_.end(); ++iter) {
	    totalBytes += iter->size();
	    totalBytes += integerSize;
	}

	// Header : package version character(5 bytes) + 
	//
	//		data total bytes(in integer - probably 4 bytes) 

	char* header = new char[5 + integerSize];
	strncpy(header, PACKAGE_VERSION, 5);
	memcpy(header + 5, reinterpret_cast<char*>(&totalBytes), integerSize);

	Send(header, 5 + integerSize);

	delete [] header;

	int processBytes = 0;
	char* buf = new char[totalBytes];

	for(std::vector<std::string>::const_iterator iter = arguments_.begin();
		iter != arguments_.end(); ++iter) {

	    // Send data : size in bytes that not contain terminator
	    //		(in integer - probably 4 bytes) +
	    //
	    //		real data
	    //
	    thisBytes = iter->size();
	    memcpy(buf + processBytes, reinterpret_cast<char*>(&thisBytes), integerSize);
	    processBytes += integerSize;

	    memcpy(buf + processBytes, iter->c_str(), iter->size());
	    processBytes += iter->size();
	}

	// Send data and receive signal
	Send(buf, totalBytes);
	RecvSignal();

	delete [] buf;
    }
}

/// @brief Start server loop.
///
/// If success on listen, execute accept loop thread.
void UnixSocketHandler::StartServerLoop()
{
    if(Listen())
	Accept();
}

/// @brief Set MainManager pointer.
void UnixSocketHandler::SetMainManager(MainManager* const mainManager)
{
    mainManager_ = mainManager;
    signalReceived_.connect(sigc::mem_fun(*mainManager_, &MainManager::SocketReceived));
}

/// @brief Process received data.
///
/// First, test data has valid header.
///
/// Header : maybe 9 bytes
///
///	    Package version charater(5 bytes) + Data total bytes(in integer - probably 4 bytes)
///
/// After, received all data to Data total bytes. And separate received data
/// to each element. Each element in received data has header for size.
///
/// Each element : size is variable
///
///	    Size in bytes(in integer - probably 4 bytes) + Real data 
///
/// @param fd Client socket
/// @param receivedBytes Received data size in first
/// @param buf Buffer to receive data
/// @param bufSize Buffer size
void UnixSocketHandler::DataReceived(const int& fd, const int& receivedBytes, char* buf, const int& bufSize)
{
    // Header : package version character(5 bytes) + 
    //
    //	    data total bytes(in integer - probably 4 bytes) 

    int mainHeaderSize = 5 + integerSize_;
    // check version in received data
    if(receivedBytes >= mainHeaderSize && (strncmp(PACKAGE_VERSION, buf, 5) == 0)) {    
	int totalBytes = 0;
	memcpy(&totalBytes, buf+5, integerSize_);

	int rcvBytes = -1;
	int curBytes = 0;
	int leftBytes = totalBytes;
	char* rcvData = new char[totalBytes];

	// If received not only header
	if(receivedBytes > mainHeaderSize) {
	    memcpy(rcvData, buf + mainHeaderSize, receivedBytes - mainHeaderSize);
	    curBytes += receivedBytes - mainHeaderSize;
	    leftBytes -= receivedBytes - mainHeaderSize;
	}

	// Received all data to expected total bytes
	while(curBytes < totalBytes) {

	    rcvBytes = recv(fd, buf, bufSize, 0);
	    // -1 is error and 0 is disconnection probably... 
	    if(rcvBytes <= 0)
		break;

	    memcpy(rcvData + curBytes, buf, rcvBytes);

	    curBytes += rcvBytes;
	    leftBytes -= rcvBytes;  

	}

	// If cound not received entire data(client disconnection ?) 
	if(leftBytes > 0)
	    totalBytes -= leftBytes;

	// I got, thanks to client!
	SendSignal(fd);

	char* eachElem;
	int eachBytes = 0;
	curBytes = 0;
	leftBytes = totalBytes;
	{
	    Glib::Mutex::Lock lock(acceptMutex_);

	    // This will isolate size part and real data part
	    while(curBytes < totalBytes) {

		// Received data : size in bytes that not contain 
		//		terminator(in integer - probably 4 bytes) +
		//		
		//		real data

		if(leftBytes < integerSize_)
		    break;

		memcpy(&eachBytes, rcvData + curBytes, integerSize_);
		curBytes += integerSize_;
		leftBytes -= integerSize_;

		if(leftBytes < eachBytes || eachBytes < 0)
		    break;

		// Need to allocate more 1 bytes for terminator
		eachElem = new char[eachBytes + 1];
		bzero(eachElem, eachBytes + 1);
		memcpy(eachElem, rcvData + curBytes, eachBytes);
		curBytes += eachBytes;
		leftBytes -= eachBytes;

		receivedData_.push_back(eachElem);

		delete [] eachElem;
	    }
	}

	delete [] rcvData;
	if(mainManager_ != NULL)
	    signalReceived_();
    }
}

