// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag_aac.cc
/// @brief Class for mpeg 4 tag processing source file.

#include "../include/tag_aac.h"
#include "../include/mainmanager.h"
#include "../include/scopeguard.h"
#include "../include/misc.h"

int AACTag::adtsSampleRate[] = {96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050, 16000, 12000, 11025, 8000, 7350, 0, 0, 0};

AACTag::AACTag(const std::string& filename, const MainManager* const mainManager)
: filename_(filename),
    process_(false),
    mp4File_(false),
    mainManager_(mainManager)
{
    FILE* inFile = std::fopen(filename.c_str(), "rb");
    ScopeGuard closeFile = MakeGuard(std::fclose, inFile);
    if(inFile == NULL)
	closeFile.Dismiss();
    else 
	mp4File_ = FAAD::AACPlayer::IsMP4File(inFile);
}

AACTag::~AACTag()
{}

void AACTag::ProcessMP4()
{
    FILE* mp4File = std::fopen(filename_.c_str(), "rb");

    if(mp4File == NULL)
	return;

    ScopeGuard closeFile = MakeGuard(std::fclose, mp4File);

    const std::auto_ptr<mp4ff_callback_t> mp4Cb(new mp4ff_callback_t);
    mp4Cb->read = MP4ReadCallback;
    mp4Cb->seek = MP4SeekCallback;
    mp4Cb->user_data = mp4File;

    faacDecHandle decoder = faacDecOpen();
    ScopeGuard closeDecoder = MakeGuard(faacDecClose, decoder);

    mp4ff_t* inFile = mp4ff_open_read(mp4Cb.get());
    ScopeGuard closeMP4ffFile = MakeGuard(mp4ff_close, inFile);
    if(inFile == NULL) {
	closeMP4ffFile.Dismiss();

	return;
    }

    int track = FAAD::AACPlayer::GetAACTrack(inFile);
    if(track < 0)
	return;

    unsigned char* buffer = NULL;
    unsigned int bufferSize = 0;	
    mp4ff_get_decoder_config(inFile, track, &buffer, &bufferSize);

    unsigned long sampleRate = 0;
    unsigned char channels = 0;
    if(faacDecInit2(decoder, buffer, bufferSize, &sampleRate, &channels) < 0)
	return;

    unsigned int frameSize = 1024;
    mp4AudioSpecificConfig mp4ASC;

    if(buffer != NULL) {
	if(AudioSpecificConfig(buffer, bufferSize, &mp4ASC) >= 0) {
	    if(mp4ASC.frameLengthFlag == 1)
		frameSize = 960;
	    if(mp4ASC.sbr_present_flag == 1)
		frameSize *= 2;	
	}
	free(buffer);
	buffer = NULL;
	bufferSize = 0;
    }

    long numSamples = mp4ff_num_samples(inFile, track);
    totalTimeSeconds_ = static_cast<unsigned long>(numSamples * (frameSize - 1.0) / mp4ASC.samplingFrequency);

    char* value = NULL;
    if(mp4ff_meta_get_title(inFile, &value)) {
	if(value != NULL) {
	    title_ = value;
	    free(value);
	    value = NULL;

	    if(StringIsSpace(title_))
		StringToUTF8(Glib::path_get_basename(filename_), "", mainManager_->GetAssumeCharset());
	    else	
		StringToUTF8(title_, Glib::path_get_basename(filename_), mainManager_->GetAssumeCharset());
	}
    }

    if(mp4ff_meta_get_artist(inFile, &value)) {
	if(value != NULL) {
	    artist_ = value;
	    free(value);
	    value = NULL;

	    StringToUTF8(artist_, "", mainManager_->GetAssumeCharset());
	}
    }

    if(mp4ff_meta_get_album(inFile, &value)) {
	if(value != NULL) {
	    album_ = value;
	    free(value);
	    value = NULL;

	    StringToUTF8(album_, "", mainManager_->GetAssumeCharset());
	}
    }

    if(mp4ff_meta_get_track(inFile, &value)) {
	if(value != NULL) {
	    track_ = value;
	    free(value);
	    value = NULL;

	    StringToUTF8(track_, "", mainManager_->GetAssumeCharset());
	}
    }

    if(mp4ff_meta_get_date(inFile, &value)) {
	if(value != NULL) {
	    year_ = value;
	    free(value);
	    value = NULL;

	    StringToUTF8(year_, "", mainManager_->GetAssumeCharset());
	}
    }

    if(mp4ff_meta_get_genre(inFile, &value)) {
	if(value != NULL) {
	    genre_ = value;
	    free(value);
	    value = NULL;

	    StringToUTF8(genre_, "", mainManager_->GetAssumeCharset());
	}
    }

    if(mp4ff_meta_get_comment(inFile, &value)) {
	if(value != NULL) {
	    comment_ = value;
	    free(value);
	    value = NULL;

	    StringToUTF8(comment_, "", mainManager_->GetAssumeCharset());
	}
    }
}

/// @todo Need testing.
void AACTag::ProcessAAC()
{
    FAAD::aac_buffer aacBuffer;
    aacBuffer.infile = std::fopen(filename_.c_str(), "rb");
    ScopeGuard closeFile = MakeGuard(std::fclose, aacBuffer.infile);
    if(aacBuffer.infile == NULL) {
	closeFile.Dismiss();

	return;
    }

    std::fseek(aacBuffer.infile, 0, SEEK_END);
    int fileRead = std::ftell(aacBuffer.infile);
    std::fseek(aacBuffer.infile, 0, SEEK_SET);

    aacBuffer.buffer = static_cast<unsigned char*>(std::malloc(FAAD_MIN_STREAMSIZE));
    if(!aacBuffer.buffer)
	return;
    std::memset(aacBuffer.buffer, 0, FAAD_MIN_STREAMSIZE);

    int bread = std::fread(aacBuffer.buffer, 1, FAAD_MIN_STREAMSIZE, aacBuffer.infile);
    aacBuffer.bytes_into_buffer = bread;
    aacBuffer.bytes_consumed = 0;
    aacBuffer.file_offset = 0;

    if(bread != FAAD_MIN_STREAMSIZE)
	aacBuffer.at_eof = 1;

    int tagSize = 0;
    if(memcmp(aacBuffer.buffer, "ID3", 3) != 0) {
	tagSize = (aacBuffer.buffer[6] << 21) | (aacBuffer.buffer[7] << 14) |
	    (aacBuffer.buffer[8] << 7) | (aacBuffer.buffer[9] << 0);

	tagSize += 10;
	FAAD::AACPlayer::AACAdvanceBuffer(&aacBuffer, tagSize);
	FAAD::AACPlayer::AACFillBuffer(&aacBuffer);
    }

    float length = 0;
    if(aacBuffer.buffer[0] == 0xFF && (aacBuffer.buffer[1] & 0xF6) == 0xF0) {
	int frames = 0;
	int frameLength = 0;
	int sampleRate = 0;
	float framesPerSec = 0.0f;
	for(;; ++frames) {
	    FAAD::AACPlayer::AACFillBuffer(&aacBuffer);
	    if(aacBuffer.bytes_into_buffer > 7) {
		if(!(aacBuffer.buffer[0] == 0xFF && (aacBuffer.buffer[1] & 0xF6) == 0XF0))
		    break;

		if(frames == 0)
		    sampleRate = adtsSampleRate[(aacBuffer.buffer[2]&0x3c)>>2];

		frameLength = ((static_cast<unsigned int>(aacBuffer.buffer[3]) & 0x3) << 11) |
		    (static_cast<unsigned int>(aacBuffer.buffer[4]) << 3) |
		    aacBuffer.buffer[5] >> 5;

		if(frameLength > aacBuffer.bytes_into_buffer)
		    break;

		FAAD::AACPlayer::AACAdvanceBuffer(&aacBuffer, frameLength);

	    } else
		break;
	}

	framesPerSec = sampleRate / 1024.0f;
	if(framesPerSec > 0.0f)
	    length = frames / framesPerSec;

    } else if(std::memcmp(aacBuffer.buffer, "ADIF", 4) == 0) {
	const int skipSize = (aacBuffer.buffer[4] & 0x80) ? 9 : 0;
	const int bitRate = (static_cast<unsigned int>(aacBuffer.buffer[4 + skipSize] & 0x0F)<<19) |
	    (static_cast<unsigned int>(aacBuffer.buffer[5 + skipSize])<<11) |
	    (static_cast<unsigned int>(aacBuffer.buffer[6 + skipSize])<<3) |
	    (static_cast<unsigned int>(aacBuffer.buffer[7 + skipSize]) & 0xE0);

	length = fileRead;
	if(length != 0.0f && bitRate != 0)
	    length = length * 8.0f / bitRate + 0.5f;
    }

    totalTimeSeconds_ = static_cast<unsigned long>(length);

    title_ = StringToUTF8(Glib::path_get_basename(filename_), "", mainManager_->GetAssumeCharset());

}

void AACTag::ProcessAll()
{
    if(mp4File_)
	ProcessMP4();
    else
	ProcessAAC();	

    process_ = true;
}

void AACTag::ProcessYear()
{
    if(year_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessTrack()
{
    if(track_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessTitle()
{
    if(title_.empty()) {
	if(!process_)
	    ProcessAll();
	else
	    title_ = StringToUTF8(Glib::path_get_basename(filename_), "", mainManager_->GetAssumeCharset());
    }
}

void AACTag::ProcessAlbum()
{
    if(album_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessGenre()
{
    if(genre_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessArtist()
{
    if(artist_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessComment()
{
    if(comment_.empty() && !process_)
	ProcessAll();
}

void AACTag::ProcessTotalTime()
{

    if(totalTime_.empty()) {

	if(totalTimeSeconds_ == 1 && !process_)
	    ProcessAll();

	totalTime_ = ConvertSecondToTimeString(totalTimeSeconds_);	

    }
}

std::auto_ptr<Tag> CreateAACTag(const std::string& filename, const MainManager* const mainManager)
{
    return std::auto_ptr<Tag>(new AACTag(filename, mainManager));
}

