// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag_mp3.cc
/// @brief Class for process mp3 tag source file.

#include <cstdlib>
#include <cerrno>

#include <strings.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>

#include <mad.h>

#include "../include/tag_mp3.h"
#include "../include/mainmanager.h"
#include "../include/misc.h"
#include "../include/scopeguard.h"
#include "../include/trace.h"

MP3Tag::MP3Tag(const std::string& filename, const MainManager* const mainManager)
: filename_(filename),
    id3Tag_(filename.c_str()),
    id3Frame_(NULL),
    id3Field_(NULL),
    xingFlagFrames_(0x00000001L),
    xingFlags_(0),
    xingFrames_(0),
    mainManager_(mainManager)
{
    Tracer::Trace trace("MP3Tag::MP3Tag");

    memset(tmpString_, 0, sizeof(tmpString_));

}

MP3Tag::~MP3Tag()
{}

/// @brief Parse XING header if have.
/// @return Success of parsing.
bool MP3Tag::XingParse(struct mad_bitptr ptr, unsigned int bitlen)
{
    xingFlags_ = 0;
    
    if(bitlen < 64 || mad_bit_read(&ptr, 32) != 
	    (('X' << 24) | ('i' << 16) | ('n' << 8) | 'g') ) {

	xingFlags_ = 0;

	return false; 
    }

    xingFlags_ = mad_bit_read(&ptr, 32);
    bitlen -= 64;

    if (xingFlags_ & xingFlagFrames_) {
	if (bitlen < 32) {

	    xingFlags_ = 0;

	    return false;
	}

	xingFrames_ = mad_bit_read(&ptr, 32);

	return true;
    }

    return false;
}

void MP3Tag::ProcessTextField(Glib::ustring& dest, const ID3_FrameID& frameID, const bool& convertUTF8, const Glib::ustring& failString)
{
    id3Frame_ = id3Tag_.Find(frameID);
    if(id3Frame_ != NULL) {
	id3Field_ = id3Frame_->GetField(ID3FN_TEXT);
	if(id3Field_ != NULL && id3Field_->Get(tmpString_, sizeof(tmpString_)) > 0) {
	    dest = tmpString_;
	
	    if(id3Field_->GetEncoding() != ID3TE_UTF8)
		dest = StringToUTF8(dest, failString, mainManager_->GetAssumeCharset());
	}
    }
}

void MP3Tag::ProcessAll()
{}

void MP3Tag::ProcessYear()
{
    ProcessTextField(year_, ID3FID_YEAR, false);
}

void MP3Tag::ProcessTrack()
{
    ProcessTextField(track_, ID3FID_TRACKNUM, false);
}

void MP3Tag::ProcessTitle()
{
    ProcessTextField(title_, ID3FID_TITLE);

    if(StringIsSpace(title_)) 
	title_ = StringToUTF8(Glib::path_get_basename(filename_), "", mainManager_->GetAssumeCharset());
}

void MP3Tag::ProcessAlbum()
{
    ProcessTextField(album_, ID3FID_ALBUM);
}

void MP3Tag::ProcessGenre()
{
    ProcessTextField(genre_, ID3FID_CONTENTTYPE, false);
    
    if(genre_[0] == '(') {
	Glib::ustring::size_type end = genre_.find(')');

	if(end <= genre_.size()) {
	    Glib::ustring genreNumberString(genre_, 1, end);
	    long genreNumber = std::strtol(genreNumberString.c_str(), NULL, 10);

	    if(!errno && genreNumber >= 0 && genreNumber < ID3_NR_OF_V1_GENRES)
		genre_ = ID3_v1_genre_description[genreNumber];
	    else
		genre_ = "Unknown"; 
	}
    }
}

void MP3Tag::ProcessArtist()
{
    ProcessTextField(artist_, ID3FID_LEADARTIST);
}

void MP3Tag::ProcessComment()
{
    ProcessTextField(comment_, ID3FID_COMMENT);
}

/// @brief Calculate total time.
void MP3Tag::ProcessTotalTime()
{
    Tracer::Trace trace("MP3Tag::ProcessTotalTime");

    int fd = open(filename_.c_str(), O_RDONLY); 
    ScopeGuard fileClose = MakeGuard(close, fd);
    
    if(fd == -1) {
	trace.Debug("Can't open file");
	fileClose.Dismiss();
	
	return;
    }

    struct stat stat;

    if(fstat(fd, &stat) == -1) {
	trace.Debug("Can't read file stat");
	
	return;	
    }

    if(!S_ISREG(stat.st_mode)) {
	trace.Debug("File isn't regular file");
	
	return;
    }

    void* fdm = mmap(0, stat.st_size, PROT_READ, MAP_SHARED, fd, 0);	
    ScopeGuard fileUnmap = MakeGuard(munmap, fdm, stat.st_size);

    if(fdm != MAP_FAILED) {
	mad_timer_t duration = mad_timer_zero;
	struct mad_stream stream;
	struct mad_header header;
	struct mad_frame frame;

	mad_stream_init(&stream);	
	mad_header_init(&header);
	mad_frame_init(&frame);
	mad_stream_buffer(&stream, static_cast<unsigned char*>(fdm), stat.st_size);

	unsigned int count = 0;
	double averageBitrate = 0.0;
	FileType fileType = VBR;

	for(;;) {
	    if(mad_header_decode(&header, &stream) == -1) {
		if(MAD_RECOVERABLE(stream.error)) 
		    continue;
		else
		    break;
	    }

	    if(count == 0) {
		// You need to use mad_frame_decode to find ancillary data
		// that xing header in.
		frame.header = header;

		if(mad_frame_decode(&frame, &stream) != -1) {
		    if(XingParse(stream.anc_ptr, stream.anc_bitlen)) {

			fileType = XING_VBR;

			break;
		    }
		}

	    }

	    // may be 5 second... ( 38 * 5 = 190)
	    if(count < 190) {
		count++;
		averageBitrate += header.bitrate;
		averageBitrate /= 2;
	    } else if(count == 190) {
		// for never enter this statement 
		count++;
		// current bitrate == average bitrate, maybe constant bitrate
		if(header.bitrate == static_cast<unsigned long>(averageBitrate)) {    
		    fileType = CBR;

		    break; 
		}
	    }

	    mad_timer_add(&duration, header.duration);
	}

	switch(fileType) {
	    case CBR:
		trace.Debug("time calcurated in CBR");
		
		// this is not accuracy, but fast and can express almost exact time.
		totalTimeSeconds_ = static_cast<unsigned long>( static_cast<double>(stat.st_size) * 8.0 / static_cast<double>(header.bitrate) );

		break;
	    
	    case VBR:
		trace.Debug("time calcurated in VBR");
		
		duration = mad_timer_abs(duration);
		totalTimeSeconds_ = duration.seconds;
	    
		break;
		
	    case XING_VBR:
		trace.Debug("time calculated in Xing header VBR");
		
		mad_timer_multiply(&header.duration, xingFrames_);
		duration = header.duration;
		duration = mad_timer_abs(duration);
		totalTimeSeconds_ = duration.seconds;
	    
		break;		
	}

	totalTime_ = ConvertSecondToTimeString(totalTimeSeconds_);

	mad_frame_finish(&frame);
	mad_header_finish(&header);
	mad_stream_finish(&stream);

    } else {
	trace.Debug("Can't map file");
	fileUnmap.Dismiss();
    }
    
}

/// @brief Create MP3Tag object.
///
/// Used in TagFactory.
std::auto_ptr<Tag> CreateId3Tag(const std::string& filename, const MainManager* const mainManager)
{
    return std::auto_ptr<Tag>(new MP3Tag(filename, mainManager));
}

