// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag_ogg.cc
/// @brief Class for process ogg tag header file.

#include "../include/tag_ogg.h"
#include "../include/mainmanager.h"
#include "../include/trace.h"

OGGTag::OGGTag(const std::string& filename, const MainManager* const mainManager)
: filename_(filename),
validFile_(true),
inFile_(NULL),
mainManager_(mainManager)
{
    Tracer::Trace trace("OGGTag::OGGTag");

    inFile_ = fopen(filename_.c_str(), "rb");
    
    if(inFile_ == NULL) {
	trace.Debug("fopen failed");
	validFile_ = false;

    }

    if(validFile_ && ov_open(inFile_, &vorbisFile_, NULL, 0) < 0) { 
	trace.Debug("ov_open failed");	

	if(fclose(inFile_) != 0)
	    trace.Debug("fclose failed");
	
	inFile_ = NULL;    
	validFile_ = false;

    }
    
    if(validFile_ && vorbisFile_.vc->comments > 0) {

	typedef std::string::iterator iter;

	std::string eachComment, key, value;
	iter equal;
	iter begin;
	iter end;
	
	for(int i = 0; i < vorbisFile_.vc->comments; ++i)
	{
	    eachComment = vorbisFile_.vc->user_comments[i];
	    begin = eachComment.begin();
	    end = eachComment.end();
	    
	    equal = find_if(begin, end, IsCharEqual());

	    if(equal != end) {
		key = std::string(begin, equal);
		// Recommendation is upper case, but it is just recommendation.
		// So convert all them to upper case.
		for_each(key.begin(), key.end(), ToUpper());
		
		equal++;
		value = std::string(equal, end);
		
		tag_.insert(std::pair<std::string, std::string>(key, value));
	    }
	}

    }
}

OGGTag::~OGGTag()
{
    Tracer::Trace trace("OGGTag::~OGGTag");
    // NOTICE:
    //	According to vorbis api documentation, if you success on ov_open,
    // you should use ov_clear and not use fclose. But you fail on ov_open,
    // you should use fclose to close a FILE pointer.

    if(validFile_) { 
	ov_clear(&vorbisFile_);
    } else if(inFile_ != NULL) {
	if(fclose(inFile_) != 0)
	    trace.Debug("fclose failed");

    }
}

void OGGTag::ProcessAll()
{

}

void OGGTag::ProcessYear()
{
    FindAndAssign("DATE", year_);
}

void OGGTag::ProcessTrack()
{
    FindAndAssign("TRACKNUMBER", track_);
}

void OGGTag::ProcessTitle()
{
    FindAndAssign("TITLE", title_, Glib::path_get_basename(filename_));
}

void OGGTag::ProcessAlbum()
{
    FindAndAssign("ALBUM", album_);
}

void OGGTag::ProcessGenre()
{
    FindAndAssign("GENRE", genre_);
}

void OGGTag::ProcessArtist()
{
    FindAndAssign("ARTIST", artist_);
}

void OGGTag::ProcessComment()
{
    FindAndAssign("COMMENT", comment_);
}

void OGGTag::ProcessTotalTime()
{
    if(validFile_) {
	totalTimeSeconds_ = static_cast<unsigned long int>(ov_time_total(&vorbisFile_, -1));

	totalTime_ = ConvertSecondToTimeString(totalTimeSeconds_);
    } 
}

/// @brief Create OGGTag object.
///
/// Used in TagFactory.
std::auto_ptr<Tag> CreateOGGTag(const std::string& filename, const MainManager* const mainManager)
{
    return std::auto_ptr<Tag>(new OGGTag(filename, mainManager));
}

