// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// wma decoder came from bmp-wma plugin.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag_wma.cc
/// @brief Class for process wma tag source file.

#include <glibmm.h>

#include "../include/tag_wma.h"
#include "../include/mainmanager.h"
#include "../include/misc.h"
#include "../include/trace.h"

WMATag::WMATag(const std::string& filename, const MainManager* const mainManager)
: filename_(filename),
validFile_(true),
avCodec_(NULL),
avCodecContext_(NULL),
avFormatContext_(NULL),
mainManager_(mainManager)
{
    validFile_ = Init();
}

WMATag::~WMATag()
{
    if(validFile_) {
	avcodec_close(avCodecContext_);
	av_close_input_file(avFormatContext_);
    }
}

bool WMATag::Init()
{
    Tracer::Trace trace("WMATag::Init");

    avcodec_init();
    avcodec_register_all();
    av_register_all();

    if(av_open_input_file(&avFormatContext_, filename_.c_str(), NULL, 0, NULL) < 0) {
	trace.Debug("av_open_input_file failed");	

	return false;
    }
    
    for(int i = 0; i < avFormatContext_->nb_streams; ++i)
    {
	avCodecContext_ = &avFormatContext_->streams[i]->codec;
	if(avCodecContext_->codec_type == CODEC_TYPE_AUDIO)
	    break;	
    }

    av_find_stream_info(avFormatContext_);

    avCodec_ = avcodec_find_decoder(avCodecContext_->codec_id);
    if(!avCodec_) {
	trace.Debug("avcodec_find_decoder failed");
		
	return false;
    }

    if(avcodec_open(avCodecContext_, avCodec_) < 0) {
	trace.Debug("avcodec_open failed");

	return false;
    }

    return true;
}

void WMATag::ProcessAll()
{

}

void WMATag::ProcessYear()
{
    if(validFile_) {
	const int year = avFormatContext_->year;
	if(year != 0) {
	    char tmp[8];
	    snprintf(tmp, 8, "%d", year);
	    year_ = tmp;
	}
    }
}

void WMATag::ProcessTrack()
{
    if(validFile_) {
	const int track = avFormatContext_->track;
	if(track != 0) {
	    char tmp[8];
	    snprintf(tmp, 8, "%d", track);
	    track_ = tmp;
	}
    }
}

void WMATag::ProcessTitle()
{
    if(validFile_) {
	if(avFormatContext_->title[0] != '\0')
	    title_ = avFormatContext_->title;
	
	if(StringIsSpace(title_))
	    title_ = StringToUTF8(Glib::path_get_basename(filename_), "", mainManager_->GetAssumeCharset());
	    
	title_ = StringToUTF8(title_, Glib::path_get_basename(filename_), mainManager_->GetAssumeCharset());    
    } else
	title_ = StringToUTF8(Glib::path_get_basename(filename_), "", mainManager_->GetAssumeCharset());
}

void WMATag::ProcessAlbum()
{
    if(validFile_) {
	if(avFormatContext_->album[0] != '\0')
	    album_ = avFormatContext_->album;
	
	album_ = StringToUTF8(album_, "", mainManager_->GetAssumeCharset());    
    }
}

void WMATag::ProcessGenre()
{
    if(validFile_) {
	if(avFormatContext_->genre[0] != '\0')
	    genre_ = avFormatContext_->genre;
	
	genre_ = StringToUTF8(genre_, "", mainManager_->GetAssumeCharset());    
    }
}

void WMATag::ProcessArtist()
{
    if(validFile_) {
	if(avFormatContext_->author[0] != '\0')
	    artist_ = avFormatContext_->author;
	
	artist_ = StringToUTF8(artist_, "", mainManager_->GetAssumeCharset());    
    }
}

void WMATag::ProcessComment()
{
    if(validFile_) {
	if(avFormatContext_->comment[0] != '\0')
	    comment_ = avFormatContext_->comment;
	
	comment_ = StringToUTF8(comment_, "", mainManager_->GetAssumeCharset());    
    }
}

void WMATag::ProcessTotalTime()
{
    if(validFile_) {
	if(avFormatContext_->duration != 0) {
	    totalTimeSeconds_ = avFormatContext_->duration/1000000LL;
	    
	    totalTime_ = ConvertSecondToTimeString(totalTimeSeconds_);
	}
    }
}

/// @brief Create WMATag object.
///
/// Used in TagFactory.
std::auto_ptr<Tag> CreateWMATag(const std::string& filename, const MainManager* const mainManager)
{
    return std::auto_ptr<Tag>(new WMATag(filename, mainManager));
}

