// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file trayicon.cc
/// @brief Class for trayicon source file

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <libintl.h>

#include <gtkmm/eventbox.h>
#include <gtkmm/image.h>
#include <gtkmm/tooltips.h>
#include <gtkmm/menu.h>
#include <gtkmm/menuitem.h>
#include <gtkmm/stock.h>

extern "C"
{

#include "../include/eggtrayicon.h"

}

#include "../include/trayicon.h"
#include "../include/mainmanager.h"

/// @brief Class for trayicon implementation
class TrayIcon::TrayIconImpl
{
    public:
	TrayIconImpl(MainManager& mainManager);
	~TrayIconImpl();

	MainManager& mainManager_;	    ///< Program control MainManager
	
	EggTrayIcon* eggTrayIcon_;	    ///< Trayicon widget in libegg
	Gtk::EventBox trayIconEventBox_;    ///< EventBox to grab event
	Gtk::Image trayIconImage_;	    ///< Image for trayicon
	Gtk::Tooltips trayIconToolTips_;    ///< Tooltips for display status

	/// @brief Gtk::Menu class to use realize() in public
	class TrayIconMenu : public Gtk::Menu
	{
	    public:
		TrayIconMenu() {}
		virtual ~TrayIconMenu() {}

		void Realize() { realize(); }
	};
	
	TrayIconMenu trayIconMenu_;	    ///< Menu for trayicon when appear 3 button clicked
	Gtk::MenuItem mainWndOpenMenu_;	    ///< MenuItem variable to control in anytime. 

	bool OnBtnPress(GdkEventButton* ev);
	bool OnScroll(GdkEventScroll* ev);
	void MenuPositionCalc(int& x, int& y, bool& pushIn);

};

TrayIcon::TrayIcon(MainManager& mainManager)
: impl_(new TrayIconImpl(mainManager))
{}

TrayIcon::~TrayIcon()
{}

/// @brief Set tooltips for display status
/// @param string String to display 
void TrayIcon::SetTooltips(const Glib::ustring& string)
{
    impl_->trayIconToolTips_.set_tip(impl_->trayIconEventBox_, string);
}

TrayIcon::TrayIconImpl::TrayIconImpl(MainManager& mainManager)
: mainManager_(mainManager),
trayIconImage_( PIXMAPSDIR "sagua-small.png" ),
mainWndOpenMenu_(gettext("_Open Window"), true)
{
    eggTrayIcon_ = egg_tray_icon_new(gettext("sagua"));
    trayIconEventBox_.add_events(Gdk::BUTTON_PRESS_MASK);
    trayIconEventBox_.signal_button_press_event().connect(
	    sigc::mem_fun(*this, &TrayIcon::TrayIconImpl::OnBtnPress) );
    trayIconEventBox_.add_events(Gdk::SCROLL_MASK);
    trayIconEventBox_.signal_scroll_event().connect(
	    sigc::mem_fun(*this, &TrayIcon::TrayIconImpl::OnScroll) );
    trayIconEventBox_.add(trayIconImage_);

    // If you use Glib::wrap to wrap to gtkmm object, you'll get warning about
    // Gtk::manage to Gtk::Window.
    // It is just warning, but i don't want to see warning. So i'll use convert to
    // Gtk+ object.
    Gtk::Widget& eventBox = trayIconEventBox_;
    gtk_container_add(GTK_CONTAINER(eggTrayIcon_), eventBox.gobj());

    mainWndOpenMenu_.signal_activate().connect(
	    sigc::mem_fun(mainManager, &MainManager::MainWndOpen));

    Gtk::Menu_Helpers::MenuList& menuList = trayIconMenu_.items();
    menuList.push_back( mainWndOpenMenu_ );
    menuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("P_rev"),
	    sigc::mem_fun(mainManager, &MainManager::PlayCtrlPrev)) );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Play"),
	    sigc::mem_fun(mainManager, &MainManager::PlayCtrlPlay)) );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Pa_use"),
	    sigc::mem_fun(mainManager, &MainManager::AudioPauseToggle)) );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Stop"),
	    sigc::mem_fun(mainManager, &MainManager::PlayCtrlStop)) );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Next"),
	    sigc::mem_fun(mainManager, &MainManager::PlayCtrlNext)) );
    menuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_File information"),
	    sigc::mem_fun(mainManager, &MainManager::FileInfo)) );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_About"),
	    sigc::mem_fun(mainManager, &MainManager::About)) );
    menuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    menuList.push_back( Gtk::Menu_Helpers::StockMenuElem(Gtk::Stock::QUIT,
	    sigc::mem_fun(mainManager, &MainManager::Quit)) );
    
    gtk_widget_show_all(GTK_WIDGET(eggTrayIcon_));
    mainWndOpenMenu_.show();
}

TrayIcon::TrayIconImpl::~TrayIconImpl()
{}

/// @brief Callback function for TrayIcon EventBox's button press signal.
///
/// If button is first, act setting in preference.
/// If button is third, pop up trayicon menu.
/// @param ev GdkEvent
bool TrayIcon::TrayIconImpl::OnBtnPress(GdkEventButton* ev)
{

    if(ev->type == GDK_BUTTON_PRESS) {

	if(ev->button ==1)
	    mainManager_.TrayIconFirstBtnClickAction();
	
	if(ev->button == 3) { 
	    trayIconMenu_.popup(
		sigc::mem_fun(*this, &TrayIcon::TrayIconImpl::MenuPositionCalc),
		ev->button, ev->time);
	}

    }
    
    return false;
}

/// @brief Callback function for TrayIcon EventBox's scroll signal.
///
/// When mouse wheeling, adjust volumn.
/// @param ev GdkEvent
bool TrayIcon::TrayIconImpl::OnScroll(GdkEventScroll* ev)
{
    if(ev->direction == GDK_SCROLL_UP)
	mainManager_.MixerVolumnIncrease();

    if(ev->direction == GDK_SCROLL_DOWN)
	mainManager_.MixerVolumnDecrease();	

    return false;	
}

/// @brief Menu position calculation when menu pop up.
/// @param x
/// @param y
/// @param pushIn
void TrayIcon::TrayIconImpl::MenuPositionCalc(int& x, int& y, bool& pushIn)
{
    Glib::RefPtr<Gdk::Screen> gdkScreen = trayIconEventBox_.get_screen();

    int screenWidth = gdkScreen->get_width();
    int screenHeight = gdkScreen->get_height();

    int width, height;
    Glib::RefPtr<Gdk::Window> gdkWindow = trayIconEventBox_.get_window();
    gdkWindow->get_size(width, height);
    
    if(!trayIconMenu_.is_realized())
    	trayIconMenu_.Realize();
    int menuWidth, menuHeight;
    trayIconMenu_.get_window()->get_size(menuWidth, menuHeight);	

    gdkWindow->get_origin(x, y);

    if(y + height < screenHeight * 0.5)
	y += height;
    else
	y -= menuHeight;
	
    if(x + menuWidth > screenWidth)
	x = screenWidth - menuWidth;
			
}

